/*  readoptions.h

    Michael Chappell - FMRIB Image Analysis Group

    Moss Zhao - IBME Quantitative Biomedical Inference (QuBIc) Group

    Copyright (C) 2015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#if !defined(ReadOptions_h)
#define ReadOptions_h

#include <fstream>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#include "utils/log.h"
#include "utils/options.h"

namespace OXASL
{
class ReadOptions
{
public:
    static ReadOptions &getInstance();
    ~ReadOptions() { delete ropt; }
    Utilities::Option<bool> help;
    Utilities::Option<bool> version;

    Utilities::Option<std::string> datafile;
    Utilities::Option<std::string> maskfile;
    Utilities::Option<int>         ntis;

    Utilities::Option<std::string> inblockform;
    Utilities::Option<std::string> inaslform;
    Utilities::Option<std::string> rpts;
    Utilities::Option<bool>        ispairs;

    Utilities::Option<bool> splitpairs;
    Utilities::Option<bool> tcdiff;
    Utilities::Option<bool> surrtcdiff;

    Utilities::Option<std::string> outblockform;
    Utilities::Option<std::string> out;

    Utilities::Option<std::string> meanout;
    Utilities::Option<std::string> splitout;

    Utilities::Option<std::string> epochout;
    Utilities::Option<int>         epochlen;
    Utilities::Option<int>         epochover;
    Utilities::Option<std::string> epochunit;

    Utilities::Option<std::string> deconvout;
    Utilities::Option<std::string> aif;

    // Extrapolate the edge of the brain to fix the artefact on the edge of the brain
    // Assumes an eroded brain
    Utilities::Option<bool> extrapolate_option;
    Utilities::Option<int>  neighbour;

    // Partial volume correction (linear regression method) parameters
    Utilities::Option<std::string> pv_gm_file;
    Utilities::Option<std::string> pv_wm_file;
    Utilities::Option<int>         kernel;

    bool parse_command_line(int argc, char **argv);

private:
    ReadOptions();
    const ReadOptions &operator=(ReadOptions &);
    ReadOptions(ReadOptions &);

    Utilities::OptionParser options;

    static ReadOptions *ropt;
};

inline ReadOptions &ReadOptions::getInstance()
{
    if (ropt == NULL)
        ropt = new ReadOptions();

    return *ropt;
}

inline ReadOptions::ReadOptions()
    :

    help(std::string("-h,--help"), false,
        std::string("display this message"),
        false, Utilities::no_argument)
    ,

    version(std::string("-v,--version"), false,
        std::string("display version identification"),
        false, Utilities::no_argument)
    ,

    //input files
    datafile(std::string("--data,--datafile"), std::string("ASL datafile"),
        std::string("data file"),
        true, Utilities::requires_argument)
    , maskfile(std::string("--mask"), std::string("maskfile\n"),
          std::string("mask"),
          false, Utilities::requires_argument)
    ,

    //input file information
    ntis(std::string("--ntis"), 0,
        std::string("Number of TIs in file"),
        true, Utilities::requires_argument)
    , inblockform(std::string("--ibf,--inblockform"), std::string("rpt"),
          std::string("Input block format:\n          rpt - blocks of measurements that include all TIs\n          tis - blocks of repeated measurements at a single TI"),
          false, Utilities::requires_argument)
    , inaslform(std::string("--iaf,--inaslform"), std::string("diff"),
          std::string("ASL data form:\n          diff - differenced data {default}\n          tc   - Tag-Control pairs\n          ct   - Control-Tag pairs\n          tcb  - Tag-Control pairs, tags and controls grouped together within block\n          ctb - Control-Tag pairs, tags and controls grouped together within block"),
          false, Utilities::requires_argument)
    , rpts(std::string("--rpts"), std::string("NULL"),
          std::string("Number of repeats at each TI as comma separated list, not required if the number of repeats is same for all TIs  (only for use with --ibf=tis)"),
          false, Utilities::requires_argument)
    ,

    ispairs(std::string("--pairs,--inpairs"), false,
        std::string("Data contains adjacent pairs of measuremnts (e.g. Tag, Control) DEPRECEATED used --iaf instead\n"),
        false, Utilities::no_argument)
    ,

    //asaq(std::string("--asaq"),false,
    //	std::string("Data is as aquired: same as --blocked --pairs"),
    //	false,Utilities::no_argument),

    // manipulation options
    splitpairs(std::string("--spairs"), false,
        std::string("Split the pairs within the data, e.g. to separate tag and control images in output"),
        false, Utilities::no_argument)
    , tcdiff(std::string("--diff"), false,
          std::string("Take the difference between the pairs, i.e. Tag control difference"),
          false, Utilities::no_argument)
    , surrtcdiff(std::string("--surrdiff"), false,
          std::string("Do surround subtraction on the pairs\n"),
          false, Utilities::no_argument)
    ,

    //basic output
    outblockform(std::string("--obf,--outblockform"), std::string("notset"),
        std::string("Output block format (for --out=):\n          rpt - blocks of measurements that include all TIs\n          tis - blocks of repeated measurements at a single TI\n          Default is same as input block format (--ibf)"),
        false, Utilities::requires_argument)
    , out(std::string("--out"), std::string("Out filename"),
          std::string("Output data file"),
          false, Utilities::requires_argument)
    ,

    // other output options
    meanout(std::string("--mean"), std::string(""),
        std::string("Output ASL data having taken mean at each TI to file"),
        false, Utilities::requires_argument)
    , splitout(std::string("--split"), std::string(""),
          std::string("Split data into separate files each each TI, specify filename root\n"),
          false, Utilities::requires_argument)
    ,

    epochout(std::string("--epoch"), std::string(""),
        std::string("Output epochs of ASL data (takes mean at each TI within the epoch)"),
        false, Utilities::requires_argument)
    , epochlen(std::string("--elen,--epochlen"), 1,
          std::string("Length of epochs in number of repeats"),
          false, Utilities::requires_argument)
    , epochover(std::string("--eol,--epochol"), 0,
          std::string("Ammount of overlap between epochs in number of repeats"),
          false, Utilities::requires_argument)
    , epochunit(std::string("--eunit,--epochunit"), std::string("rpt"),
          std::string("Epochs to be determined over:\n          rpt - repeats in the data {default}\n          tis - TIs in the data\n"),
          false, Utilities::requires_argument)
    ,

    deconvout(std::string("--deconv"), std::string(""),
        std::string("Deconvolution of data with arterial input functions"),
        false, Utilities::requires_argument)
    , aif(std::string("--aif"), std::string(""),
          std::string("Arterial input functions for deconvolution (4D volume, one aif for each voxel within mask)\n"),
          false, Utilities::requires_argument)
    ,

    // Extrapolate the edge of the brain to fix the artefact on the edge of the brain
    // Assumes an eroded brain
    extrapolate_option(std::string("--extrapolate"), false,
        std::string("Option to extrapolate the edge of the brain to fix the artefact on the edge of the brain"),
        false, Utilities::no_argument)
    , neighbour(std::string("--neighbour"), 5, std::string("Neighbour size for extrapolation, must be an odd number between 3 and 9. Default: 5\n"),
          false, Utilities::requires_argument)
    ,

    // Partial volume (linear regression) options
    pv_gm_file(std::string("--pvgm"), std::string(""), std::string("GM partial volume map"),
        false, Utilities::requires_argument)
    , pv_wm_file(std::string("--pvwm"), std::string(""), std::string("WM partial volume map"),
          false, Utilities::requires_argument)
    , kernel(std::string("--kernel"), 5, std::string("Kernel size (in voxels) of partial volume correction, must be an odd number between 3 and 9. Default: 5\n"),
          false, Utilities::requires_argument)
    ,

    options("asl_file", "asl_file --data=<asldata> --ibf=rpt --iaf=tc --diff --out=<diffdata>\n")
{
    try
    {
        options.add(help);
        options.add(version);

        options.add(datafile);
        options.add(maskfile);

        options.add(ntis);
        options.add(inblockform);
        options.add(inaslform);
        options.add(rpts);
        options.add(ispairs);

        options.add(splitpairs);
        options.add(tcdiff);
        options.add(surrtcdiff);

        options.add(extrapolate_option);
        options.add(neighbour);

        options.add(pv_gm_file);
        options.add(pv_wm_file);
        options.add(kernel);

        options.add(out);
        options.add(outblockform);

        options.add(meanout);
        options.add(splitout);

        options.add(epochout);
        options.add(epochlen);
        options.add(epochover);
        options.add(epochunit);

        options.add(deconvout);
        options.add(aif);
    }

    catch (Utilities::X_OptionError &e)
    {
        options.usage();
        std::cerr << std::endl
             << e.what() << std::endl;
    }
    catch (std::exception &e)
    {
        std::cerr << e.what() << std::endl;
    }
}
}
#endif
